#!/usr/bin/python
# -*- coding: utf-8 -*-

## ************************************************************************** ##
##
##  BornAgain: simulate and fit reflection and scattering
##
##! @file      test_performance.py
##! @brief     Defines class TestPerformance for logging performance changes
##!            For explanations, see the README file
##!
##! @homepage  http:##www.bornagainproject.org
##! @license   GNU General Public License v3 or higher (see COPYING)
##! @copyright Forschungszentrum Jülich GmbH 2016
##! @authors   Scientific Computing Group at MLZ Garching
##! @authors   C. Durniak, M. Ganeva, G. Pospelov, W. Van Herck, J. Wuttke
##
## ************************************************************************** ##

from __future__ import print_function
import datetime
import sys
import platform
import timeit
from collections import OrderedDict
import math
from libBornAgainCore import *

# used for logging intermediate output
logfile = sys.stdout

# for code readability
get_wall_time = timeit.default_timer

# for code readability and portability, since resource is not guaranteed to be available
try:
    import resource
    record_cpu_time = True
    get_cpu_time = lambda: resource.getrusage(resource.RUSAGE_SELF)[0]
except:
    record_cpu_time = False
    get_cpu_time = lambda: None

# globals used in custom form factor
phi_min, phi_max = -1, 1.0
alpha_min, alpha_max = 0, 2.0


# user-defined custom form factor
class CustomFormFactor(IBornFF):
    """
    A custom defined form factor
    The form factor is V sech(q L) with
    V volume of particle
    L length scale which defines mean radius
    """
    def __init__(self, V, L):
        IBornFF.__init__(self)
        # parameters describing the form factor
        self.V = V
        self.L = L

    def clone(self):
        """
        IMPORTANT NOTE:
        The clone method needs to call transferToCPP() on the cloned object
        to transfer the ownership of the clone to the cpp code
        """
        cloned_ff = CustomFormFactor(self.V, self.L)
        cloned_ff.transferToCPP()
        return cloned_ff

    def evaluate_for_q(self, q):
        return self.V*1.0/math.cosh(q.mag()*self.L)


# class for performance test, constructed using sample factories
class FactoryTest:
    def __init__(self, name, simulation_name, sample_builder,
                 nrepetitions):
        self.m_test_name = name
        self.m_simulation_name = simulation_name
        self.m_sample_builder_name = sample_builder
        self.m_nrepetitions = nrepetitions
        self.m_cpu_time = 0.0
        self.m_wall_time = 0.0

        self.m_sample = None

        if simulation_name != None and sample_builder != None:
            self.m_sample_factory = SampleBuilderFactory()
            self.m_simulation_factory = SimulationFactory()
            self.m_simulation = self.m_simulation_factory.createItem(
                self.m_simulation_name)
            self.m_sample = self.m_sample_factory.createSampleByName(
                self.m_sample_builder_name)
        else:
            self.m_sample_factory = None
            self.m_simulation_factory = None
            self.m_simulation = None
            self.m_sample = None

    def prepare(self):
        pass

    # do the actual work
    def run_loop(self):
        # actual work is done here
        for i in range(self.m_nrepetitions):
            self.m_simulation.setSample(self.m_sample)
            self.m_simulation.runSimulation()

    def run(self):

        self.prepare()

        logfile.write("Running test: %-30s " % self.m_test_name)
        logfile.flush()

        # resource.getrusage is less accurate but returns user (cpu) time
        start_wall_time = get_wall_time()
        start_cpu_time = get_cpu_time()

        self.run_loop()

        end_wall_time = get_wall_time()
        end_cpu_time = get_cpu_time()

        # note that on multi-core system, cpu time can be higher than wall time
        self.m_wall_time = end_wall_time - start_wall_time

        if record_cpu_time:
            self.m_cpu_time = end_cpu_time - start_cpu_time
            logfile.write("OK: %-6.3f (wall sec), %-6.3f (cpu sec) \n" %
                          (self.m_wall_time, self.m_cpu_time))
        else:
            logfile.write("OK: %-6.3f (wall sec)\n" % (self.m_wall_time))


# special performance test case: custom form factor
class CustomTest(FactoryTest):
    def __init__(self, name, nrepetitions):
        FactoryTest.__init__(self, name, None, None, nrepetitions)
        self.m_sample = None
        self.m_simulation = None

    def prepare(self):
        self.m_sample = self.get_sample()
        self.m_simulation = self.get_simulation()

    # do the actual work
    def run_loop(self):
        for i in range(self.m_nrepetitions):
            self.m_simulation.setSample(self.m_sample)
            self.m_simulation.runSimulation()

    def get_sample(self):
        """
        Build and return the sample to calculate custom form factor in Distorted Wave Born Approximation.
        """
        # defining materials
        m_vacuum = HomogeneousMaterial("Vacuum", 0, 0)
        m_substrate = HomogeneousMaterial("Substrate", 6e-6, 2e-8)
        m_particle = HomogeneousMaterial("Particle", 6e-4, 2e-8)

        # collection of particles
        ff = CustomFormFactor(343*nm, 7*nm)
        particle = Particle(m_particle, ff)
        particle_layout = ParticleLayout()
        particle_layout.addParticle(particle, 1)
        vacuum_layer = Layer(m_vacuum)
        vacuum_layer.addLayout(particle_layout)
        substrate_layer = Layer(m_substrate)

        # assemble multilayer
        multi_layer = MultiLayer()
        multi_layer.addLayer(vacuum_layer)
        multi_layer.addLayer(substrate_layer)
        return multi_layer

    def get_simulation(self):
        """
        Create and return GISAXS simulation with beam and detector defined
        IMPORTANT NOTE:
        Multithreading should be deactivated by putting ThreadInfo.n_threads to -1
        """
        simulation = GISASSimulation()
        simulation.getOptions().setNumberOfThreads(-1)
        simulation.setDetectorParameters(100, phi_min*deg, phi_max*deg,
                                         100, alpha_min*deg, alpha_max*deg)
        simulation.setBeamParameters(1*angstrom, 0.2*deg, 0)
        return simulation


# class for running all of the tests and collecting results
class PerformanceTests:
    def __init__(self, filename=None):
        self.m_tests = []
        self.m_datime = ""
        self.m_hostname = ""
        self.m_sysinfo = ""
        self.m_pyversion = ""
        self.m_filename = filename

        self.add("MultiLayer", "MaxiGISAS",
                 "MultiLayerWithRoughnessBuilder", 1)
        self.add("CylindersInDWBA", "MaxiGISAS", "CylindersInDWBABuilder",
                 10)
        self.add("RotatedPyramids", "MaxiGISAS", "RotatedPyramidsBuilder",
                 10)
        self.add("CoreShell", "MaxiGISAS", "CoreShellParticleBuilder", 10)
        self.add("SquareLattice2D", "MaxiGISAS", "SquareLattice2DBuilder",
                 10)
        self.add("RadialParaCrystal", "MaxiGISAS",
                 "RadialParaCrystalBuilder", 10)
        self.add("HexParaCrystal", "BasicGISAS", "HexParaCrystalBuilder",
                 1)
        self.add("SSCA", "MaxiGISAS", "SizeDistributionSSCAModelBuilder",
                 10)
        self.add("Mesocrystal", "MaxiGISAS", "MesoCrystalBuilder", 2)
        self.add("PolMagCyl", "MaxiGISAS00", "MagneticCylindersBuilder",
                 10)

        # custom form factor is a special case since it's not in the registry
        self.m_tests.append(CustomTest("Custom FF", 10))

        logfile.write("\nPreparing to run %d performance tests.\n\n" %
                      len(self.m_tests))

    def add(self, name, simulation_name, sample_builder, nrepetitions):
        self.m_tests.append(
            FactoryTest(name, simulation_name, sample_builder,
                        nrepetitions))

    # execute all performance tests
    def execute(self):
        self.init_sysinfo()
        for test in self.m_tests:
            test.run()
        self.write_results()

    # write out system information and test results to file
    def write_results(self):

        if (self.m_filename != None):
            try:
                write_file = open(self.m_filename, "a")
            except:
                sys.stderr.write(
                    "Could not open filed '%' for writing. Writing to stdout instead.\n"
                    % self.m_filename)
                write_file = sys.stdout
                self.m_filename = None

        else:
            write_file = sys.stdout

        logfile.write("\nWriting output to %s...\n\n" % write_file.name)

        # record results into an ordered dict, which is used by pretty_write() below
        dictionary = OrderedDict()

        dictionary["date"] = self.m_datime
        dictionary["hostname"] = self.m_hostname
        dictionary["sysinfo"] = self.m_sysinfo
        dictionary["python"] = self.m_pyversion

        sum_wall = 0.0
        sum_cpu = 0.0

        for test in self.m_tests:
            sum_wall += test.m_wall_time
            sum_cpu += test.m_cpu_time

        if record_cpu_time: dictionary["total cpu"] = "%-.4f" % sum_cpu
        dictionary["total wall"] = "%-.4f" % sum_wall

        for test in self.m_tests:
            dictionary[test.m_test_name] = "%-.4f" % test.m_wall_time

        write_file.write("\n")
        pretty_write(dictionary, write_file)
        write_file.write("\n")
        write_file.flush()

        if (self.m_filename != None):
            write_file.close()

    # determine platform, architecture, python version, etc.
    def init_sysinfo(self):
        system, node, release, version, machine, processor = platform.uname(
        )
        self.m_datime = datetime.datetime.strftime(datetime.datetime.now(),
                                                   '%Y-%m-%d %H:%M:%S')
        self.m_hostname = node
        self.m_sysinfo = "%s %s" % (system, machine)
        self.m_pyversion = "%d.%d" % (sys.version_info[:2])


# used for writing the summary with proper formatting
def pretty_write(dictionary, file):
    header = "|"
    footer = "|"

    for key in dictionary:
        value = dictionary[key]
        length = max(len(key), len(value))
        format_str = " {0:" + str(length) + "} |"
        header += format_str.format(key)
        footer += format_str.format(value)

    file.write(header + "\n")
    file.write(footer + "\n")


if __name__ == '__main__':

    if (len(sys.argv) > 2):
        print("Usage: test_performance.py [output file]")
        exit(0)

    if (len(sys.argv) == 2):
        filename = sys.argv[1]
    else:
        filename = None

    tests = PerformanceTests(filename)
    tests.execute()



#!/usr/bin/env python3
# Draws two dimensional numpy array using matplotlib
# Usage: python show2d.py file_name

from __future__ import print_function
from pylab import *
from matplotlib.colors import LogNorm
import argparse
import os

# Define a nice colormap
cdict = {
    'red': ((0.0, 0.0, 0.0), (0.5, 0.0, 0.0), (1.0, 1.0, 1.0)),
    'green': ((0.0, 0.0, 0.0), (0.5, 0.0, 0.0), (1.0, 1.0, 1.0)),
    'blue': ((0.0, 0.0, 0.0), (0.5, 1.0, 1.0), (1.0, 1.0, 1.0))
}
blue_cmap = matplotlib.colors.LinearSegmentedColormap('blue_map', cdict, 256)


def PlotNumpyArray(a, zmin=1, zmax=None):
    # Make a plot of the data
    #print( "plotting..." )
    amax = a.flatten().max()
    amean = np.sum(a)/a.size
    aminplot = amean**2/amax
    a = np.maximum(a, zmin)
    dataarray = np.flipud(np.transpose(a))
    plt.xlabel(r'$\phi_f$', fontsize=20)
    plt.ylabel(r'$\alpha_f$', fontsize=20)
    # Use one of the predefined colormaps or the above defined 'blue_cmap':
    im = plt.imshow(dataarray, norm=LogNorm(), vmax=zmax, cmap=cm.jet)
    plt.gca().axes.get_xaxis().set_ticks([])
    plt.gca().axes.get_yaxis().set_ticks([])
    plt.colorbar(im)

    # Show the plot
    #print( "showing..." )
    # Uncomment the next lines to generate a pdf from the plot
    #extension = 'pdf'
    #plt.savefig(file_no_prefix + '.%s' %extension, format=extension)
    plt.show()


#-------------------------------------------------------------
# main()
#-------------------------------------------------------------
if __name__ == '__main__':
    # Define the arguments to the script
    parser = argparse.ArgumentParser(
        description='Plot 2D data table on log scale.')
    parser.add_argument('-m',
                        '--minz',
                        dest='zmin',
                        metavar='MINZ',
                        help='minimum z-value to display [default: 1]',
                        default='1')
    parser.add_argument(
        '-M',
        '--maxz',
        dest='zmax',
        metavar='MAXZ',
        help='maximum z-value to display [default: maximum value of data set]',
        default=None)
    parser.add_argument('filename',
                        metavar='filename',
                        help='filename of the data file')

    # Parse the script's arguments
    args = parser.parse_args()
    zmin = float(args.zmin)
    if (args.zmax):
        zmax = float(args.zmax)
    else:
        zmax = None
    filename = args.filename

    # Load the file's data
    print("loading...")
    a = np.loadtxt(filename)
    file_no_prefix = os.path.splitext(os.path.basename(filename))[0]
    print('Filename: ', file_no_prefix)
    print('Data shape: ', a.shape)
    print('Minimum value of data: ', a.flatten().min())
    print('Maximum value of data: ', a.flatten().max())

    PlotNumpyArray(a, zmin, zmax)



#!/usr/bin/env python3
"""
Basic example of depth-probe simulation with BornAgain.
Sample structure:
----------------------- inf

                        Si

----------------------- 0 nm
                        Ti
----------------------- -13 nm
                        Pt
----------------------- -45 nm
                        Ti
----------------------- -55 nm
                        TiO2
----------------------- -58 nm

                        D2O

----------------------- -inf
Beam comes from silicon side.
z axis is directed up and perpendicularly
to the sample, z = 0 corresponding to the sample
surface
"""
import bornagain as ba
from bornagain import angstrom, deg, nm, nm2, kvector_t

# layer thicknesses in angstroms
t_Ti = 130*angstrom
t_Pt = 320*angstrom
t_Ti_top = 100*angstrom
t_TiO2 = 30*angstrom

#  beam data
ai_min = 0  # minimum incident angle
ai_max = 1*deg  # maximum incident angle
n_ai_bins = 5000  # number of bins in incident angle axis
beam_sample_ratio = 0.01  # beam-to-sample size ratio
wl = 10*angstrom  # wavelength in angstroms

# convolution parameters
d_ang = 0.01*ba.deg  # spread width for incident angle
n_points = 25  # number of points to convolve over
n_sig = 3  # number of sigmas to convolve over

#  depth position span
z_min = -100*nm
z_max = 100*nm
n_z_bins = 500


def get_sample():
    """
    Constructs a sample with one resonating Ti/Pt layer
    """

    # Define materials
    material_D2O = ba.HomogeneousMaterial("D2O", 0.00010116, 1.809e-12)
    material_Pt = ba.HomogeneousMaterial("Pt", 0.00010117, 3.01822e-08)
    material_Si = ba.HomogeneousMaterial("Si", 3.3009e-05, 0)
    material_Ti = ba.HomogeneousMaterial("Ti", -3.0637e-05, 1.5278e-08)
    material_TiO2 = ba.HomogeneousMaterial("TiO2", 4.1921e-05, 8.1293e-09)

    # Define layers
    layer_1 = ba.Layer(material_Si)
    layer_2 = ba.Layer(material_Ti, 13*nm)
    layer_3 = ba.Layer(material_Pt, 32*nm)
    layer_4 = ba.Layer(material_Ti, 10*nm)
    layer_5 = ba.Layer(material_TiO2, 3*nm)
    layer_6 = ba.Layer(material_D2O)

    # Define sample
    sample = ba.MultiLayer()
    sample.addLayer(layer_1)
    sample.addLayer(layer_2)
    sample.addLayer(layer_3)
    sample.addLayer(layer_4)
    sample.addLayer(layer_5)
    sample.addLayer(layer_6)

    return sample


def get_simulation(sample):
    """
    Returns a depth-probe simulation.
    """
    alpha_distr = ba.DistributionGaussian(0, d_ang)
    footprint = ba.FootprintSquare(beam_sample_ratio)
    simulation = ba.DepthProbeSimulation()
    simulation.setBeamParameters(wl, n_ai_bins, ai_min, ai_max, footprint)
    simulation.setZSpan(n_z_bins, z_min, z_max)
    simulation.addParameterDistribution("*/Beam/InclinationAngle",
                                        alpha_distr, n_points, n_sig)
    simulation.setSample(sample)
    return simulation


if __name__ == '__main__':
    import ba_plot
    sample = get_sample()
    simulation = get_simulation(sample)
    ba_plot.run_and_plot(simulation, aspect='auto')



#!/usr/bin/env python3
"""
Basic example of a DWBA simulation of a GISAS experiment:
Scattering by a dilute random assembly of cylinders on a substrate.
"""
import bornagain as ba
from bornagain import deg, nm


def get_sample():
    # Define materials
    material_Particle = ba.HomogeneousMaterial("Particle", 6e-4, 2e-08)
    material_Substrate = ba.HomogeneousMaterial("Substrate", 6e-6, 2e-08)
    material_Vacuum = ba.HomogeneousMaterial("Vacuum", 0, 0)

    # Define particles
    ff = ba.FormFactorCylinder(5*nm, 5*nm)
    particle = ba.Particle(material_Particle, ff)

    # Define particle layouts
    layout = ba.ParticleLayout()
    layout.addParticle(particle)
    layout.setTotalParticleSurfaceDensity(0.01)

    # Define layers
    layer_1 = ba.Layer(material_Vacuum)
    layer_1.addLayout(layout)
    layer_2 = ba.Layer(material_Substrate)

    # Define sample
    sample = ba.MultiLayer()
    sample.addLayer(layer_1)
    sample.addLayer(layer_2)

    return sample


def get_simulation(sample):
    # Define beam
    wavelength = 0.1*nm
    alpha_i = 0.2*deg
    beam = ba.Beam(1, wavelength, ba.Direction(alpha_i, 0))

    # Define detector
    nPix = 200  # pixels per direction
    detector = ba.SphericalDetector(nPix, -2*deg, 2*deg, nPix, 0, 3*deg)

    return ba.GISASSimulation(beam, sample, detector)


if __name__ == '__main__':
    import ba_plot
    sample = get_sample()
    simulation = get_simulation(sample)
    ba_plot.run_and_plot(simulation)



"""
Script to fix all dependencies in OS X bundle.
Runs automatically when -DBORNAGAIN_APPLE_BUNDLE.
Called via cmake/configurables/FixAppleBundle.cmake.in.
"""
from __future__ import print_function
import os
import sys
import platform
import shutil
import subprocess
import glob
from distutils import sysconfig

BUNDLE_DIR = ""

# -----------------------------------------------------------------------------
# Locations
# -----------------------------------------------------------------------------


def is_python3():
    if (sys.version_info > (3, 0)):
        return True
    else:
        return False


def python_version_string():
    return str(sys.version_info[0]) + "." + str(sys.version_info[1])


def set_bundle_dir(bundle_dir):
    global BUNDLE_DIR
    BUNDLE_DIR = bundle_dir


def bundle_dir():
    global BUNDLE_DIR
    return BUNDLE_DIR


def bundle_frameworks_path():
    return os.path.join(bundle_dir(), "Contents", "Frameworks")


def bundle_plugins_path():
    return os.path.join(bundle_dir(), "Contents", "PlugIns")


def bundle_main_executables():
    return [os.path.join(bundle_dir(), "Contents", "MacOS", "BornAgain")]


def bundle_python_library():
    return os.path.join("Python.framework", "Versions",
                        python_version_string(), "Python")


def qtlibs_path():
    return os.path.join(os.environ['QTDIR'], "lib")


def qtplugins_path():
    return os.path.join(os.environ['QTDIR'], "plugins")


def bundle_libraries():
    return glob.glob(
        os.path.join(bundle_dir(), "Contents", "lib", "BornAgain-*", "*"))


def bundle_plugins():
    return glob.glob(os.path.join(bundle_plugins_path(), "*", "*"))


def bornagain_binaries():
    return bundle_main_executables() + bundle_libraries() + bundle_plugins(
    )


def get_list_of_files(dirname):
    """
    returns recursive list of files in given directory and its subdirectories
    """
    result = []
    for root, directories, filenames in os.walk(dirname):
        for filename in filenames:
            result.append(os.path.join(root, filename))
    return result


# -----------------------------------------------------------------------------
# Utils
# -----------------------------------------------------------------------------


def make_dir(dirname):
    if not os.path.exists(dirname):
        os.makedirs(dirname)


def copy_file_to_dir(file_name, dest_dir):
    """
    Copies file to the destination directory. If destination doesn't exists, it will be created.
    """
    make_dir(dest_dir)
    dest_file = os.path.join(dest_dir, os.path.basename(file_name))
    shutil.copyfile(file_name, dest_file)
    return dest_file


def copy_file_to_file(source, destination):
    """
    Copies file to another. If destination directory doesn't exists, it will be created.
    """
    make_dir(os.path.dirname(destination))
    shutil.copyfile(source, destination)


def otool(filename):
    """
    Parses dependencies of given binary file
    """
    p = subprocess.Popen(['otool', '-XL', filename],
                         stdout=subprocess.PIPE)
    # return iter(p.stdout.readline, b'')
    for line in iter(p.stdout.readline, b''):
        if is_python3():
            yield line.decode('ascii').strip().split()[0]
        else:
            yield line.strip().split()[0]


def setId(filename, newId):
    """
    Sets new id for binary file
    """
    p = subprocess.Popen(['install_name_tool', '-id', newId, filename],
                         stdout=subprocess.PIPE)
    p.communicate()
    return


def fixDependency(filename, old, new):
    """
    Replaces old dependency with new one for given binary file
    """
    print("    fixDependency(filename, old, new)", filename, old, new)
    p = subprocess.Popen(
        ['install_name_tool', '-change', old, new, filename],
        stdout=subprocess.PIPE)
    p.communicate()
    return


def is_system_dependency(dependency):
    """
    Returns True if this is system file dependency
    """
    non_dependency_patterns = ['/usr/lib', '/System']
    for pat in non_dependency_patterns:
        if dependency.startswith(pat):
            return True
    return False


def is_rpath_dependency(dependency):
    """
    Returns True if this is rpath like dependency
    """
    non_dependency_patterns = ['@']
    for pat in non_dependency_patterns:
        if dependency.startswith(pat):
            return True
    return False


def is_python_framework_dependency(dependency):
    """
    Returns True if this dependency is python library
    """
    if not "boost" in dependency.lower():
        if "python" in dependency.lower():
            return True
    return False


def is_qt_framework_dependency(dependency):
    """
    Returns True if this dependency is Qt related
    """
    if "Qt" in dependency:
        return True
    return False


def is_to_bundle_dependency(dependency):
    """
    Returns True if this dependency should be moved to the bundle.
    Qt libraries and Python framework are special case and will be treated separately.
    """

    if is_system_dependency(dependency):
        return False

    if "libBornAgain" in dependency:
        # our own libraries are already in place and have right libId's
        return False

    return True


def get_dependency_libId(dependency):
    """
    Returns libId of this dependency
    """
    libname = os.path.basename(dependency)

    if is_python_framework_dependency(dependency):
        return "@rpath/" + bundle_python_library()

    if is_qt_framework_dependency(dependency):
        return "@rpath/" + libname + ".framework/Versions/5/" + libname

    # all other libraries
    return "@rpath/" + libname


def get_dependency_orig_location(dependency):
    """
    Returns dependency original location. In most cases it is just "dependency" itself (i.e. path to file).
    If dependency contains @rpath, then we are trying to guess where original files might be
    """

    if os.path.exists(dependency):
        return dependency

    result = None

    if is_python_framework_dependency(dependency):
        result = get_python_library_location()

    elif is_qt_framework_dependency(dependency):
        libname = os.path.basename(dependency)
        libpath = os.path.join(libname + ".framework", "Versions", "5")
        result = os.path.join(qtlibs_path(), libpath, libname)

    return result


def get_dependency_dest_location(dependency):
    """
    Return new location of the dependency in the bundle
    """
    libname = os.path.basename(dependency)

    if is_python_framework_dependency(dependency):
        return os.path.join(bundle_frameworks_path(),
                            bundle_python_library())

    if is_qt_framework_dependency(dependency):
        libpath = os.path.join(libname + ".framework", "Versions", "5")
        return os.path.join(bundle_frameworks_path(), libpath, libname)

    return os.path.join(bundle_frameworks_path(), libname)


def get_special_dependency_id(dependency):
    """
    Returns libId if this dependency requires special attention (i.e. Python framework)
    """
    if is_python_framework_dependency(dependency):
        return "@rpath/" + bundle_python_library()

    if is_qt_framework_dependency(
            dependency) and not "@rpath" in dependency:
        libname = os.path.basename(dependency)
        return "@rpath/" + libname + ".framework/Versions/5/" + libname
    return None


def get_python_library_location():
    """
    Returns location of Python library. The library is deduced from interpreter itself
    """
    for dependency in otool(sys.executable):
        print(dependency)
        if os.path.exists(dependency) and "Python.framework" in dependency:
            return dependency

    # At this point it looks that interpreter depends from libPython via @loader_path
    # Let's try to find library directly

    prefix = sys.prefix
    suffix = sysconfig.get_config_var(
        'LDVERSION') or sysconfig.get_config_var('VERSION')
    result = sys.prefix + "/lib/libpython" + suffix + ".dylib"
    if os.path.exists(result):
        return result

    return None


# -----------------------------------------------------------------------------
# Actions
# -----------------------------------------------------------------------------


def copy_python_framework():
    """
    Copies Python library to the bundle. The name of the library will be deduced from the interpreter itself.
    """
    print("--> Copying Python framework")
    python_lib = get_python_library_location()
    destfile = os.path.join(bundle_frameworks_path(),
                            bundle_python_library())
    make_dir(os.path.dirname(destfile))
    print("    From '{0}'\n    To '{1}'".format(python_lib, destfile))
    if not os.path.exists(destfile):
        shutil.copyfile(python_lib, destfile)
        libId = "@rpath/" + bundle_python_library()
        setId(destfile, libId)


def copy_qt_libraries():
    print("--> Copying Qt libraries")
    libs = [
        'QtCore', 'QtDBus', 'QtDesigner', 'QtGui', 'QtPrintSupport',
        'QtWidgets', 'QtXml', 'QtSvg', 'QtNetwork', 'QtOpenGL'
    ]
    print("   ", end="")
    for libname in libs:
        print(libname, end="")
        libpath = os.path.join(libname + ".framework", "Versions", "5")
        srcfile = os.path.join(qtlibs_path(), libpath, libname)
        if os.path.exists(srcfile):
            dstdir = os.path.join(bundle_frameworks_path(), libpath)
            copy_file_to_dir(srcfile, dstdir)


def copy_qt_plugins():
    print("--> Copying Qt plugins")
    plugins = [
        'platforms/libqcocoa.dylib', 'iconengines/libqsvgicon.dylib',
        'imageformats/libqjpeg.dylib', 'imageformats/libqsvg.dylib',
        'styles/libqmacstyle.dylib'
    ]
    print("   ", end="")
    for name in plugins:
        print(name, end="")
        srcfile = os.path.join(qtplugins_path(), name)
        dstdir = os.path.join(bundle_plugins_path(), os.path.dirname(name))
        dstfile = copy_file_to_dir(srcfile, dstdir)
        setId(dstfile, os.path.basename(dstfile))


def process_dependency(dependency):
    """
    Copies external dependency library into Frameworks directory (if it is not already there)
    and sets corresponding library Id for it. Returns library Id and new location to the user.
    """

    libId = get_dependency_libId(dependency)
    origLocation = get_dependency_orig_location(dependency)
    destLocation = get_dependency_dest_location(dependency)

    print("     ------")
    if libId == dependency:
        print("     selfDependency")
        return None, None

    if origLocation == None:
        print("     origLocation >", origLocation)
        return libId, None

    print("     origLocation >", origLocation)
    print("     destLocation >", destLocation)
    print("            libId >", libId)

    if os.path.exists(destLocation):
        return libId, None
    print("       copying to >", destLocation)
    copy_file_to_file(origLocation, destLocation)
    setId(destLocation, libId)
    return libId, destLocation


def walk_through_dependencies(file_name):
    print("============================")
    print("walk_through ", file_name)
    for dependency in otool(file_name):
        print("---> ", file_name, dependency)
        if is_to_bundle_dependency(dependency):
            libId, new_location = process_dependency(dependency)
            print("to Bundle", libId, new_location)
            if libId:
                fixDependency(file_name, dependency, libId)
                if new_location:
                    walk_through_dependencies(new_location)


def copy_dependencies():
    print("--> Copying third party dependencies")
    for binfile in iter(bornagain_binaries()):
        walk_through_dependencies(binfile)


def validate_dependencies():
    """
    Analyse whole bundle for missed dependencies
    """
    binaries = bornagain_binaries()
    libraries = get_list_of_files(bundle_frameworks_path())
    file_list = binaries + libraries
    files_with_missed_dependencies = []
    for file_name in file_list:
        for dependency in otool(file_name):
            if is_system_dependency(dependency) or is_rpath_dependency(
                    dependency):
                continue
            if os.path.basename(file_name) in dependency:
                # self dependency (libId)
                continue
            files_with_missed_dependencies.append(file_name)
            break

    if len(files_with_missed_dependencies):
        print("Error! Still unresolved dependencies.")
        print(files_with_missed_dependencies)
        raise Exception("Unresolved dependencies!")


def fix_apple_bundle():
    print('-'*80)
    print("Fixing OS X bundle at '{0}'".format(bundle_dir()))
    print('-'*80)
    # # copy_python_framework()
    # FIXME provide automatic recognition of Qt dependency type (@rpath or hard coded)
    # copy_qt_libraries() # this line should be uncommented for macport based builds
    copy_qt_plugins()
    copy_dependencies()
    validate_dependencies()
    print("Done!")


if __name__ == '__main__':
    if not platform.system() == 'Darwin':
        exit("This script is intended for MacOS. Exiting...")

    if len(sys.argv) != 2:
        exit("Please specify bundle location")

    set_bundle_dir(sys.argv[1])

    fix_apple_bundle()



#!/usr/bin/env python3
"""
Plot form factor.
"""
import bornagain as ba
from bornagain import nm, deg
import bornplot as bp

det = ba.SphericalDetector(200, 5*deg, 2.5*deg, 2.5*deg)
pars = [0, 5, 10, 20]
n = len(pars)
results = []
for i in range(n):
    theta = pars[i]
    title = r'$\vartheta=%d^\circ$' % theta
    ff = ba.FormFactorTruncatedSphere(4.2*nm, 6.1*nm, 0)
    trafo = ba.RotationY(theta*deg)
    data = bp.run_simulation(det, ff, trafo)
    results.append(bp.Result(i, data, title))

bp.make_plot(results, det, "ff_demo_1quadrants")



#!/usr/bin/env python3
"""
External minimize: using lmfit minimizers for BornAgain fits.
Fit progress is plotted using lmfit iteration calbback function.
"""
import numpy as np
from matplotlib import pyplot as plt
import bornagain as ba
import ba_fitmonitor
from bornagain import deg, angstrom, nm
import lmfit


def get_sample(params):
    """
    Returns a sample with cylinders and pyramids on a substrate,
    forming a hexagonal lattice.
    """
    radius = params['radius']
    lattice_length = params['length']

    m_vacuum = ba.HomogeneousMaterial("Vacuum", 0, 0)
    m_substrate = ba.HomogeneousMaterial("Substrate", 6e-6, 2e-8)
    m_particle = ba.HomogeneousMaterial("Particle", 6e-4, 2e-8)

    sphere_ff = ba.FormFactorFullSphere(radius)
    sphere = ba.Particle(m_particle, sphere_ff)
    particle_layout = ba.ParticleLayout()
    particle_layout.addParticle(sphere)

    interference = ba.InterferenceFunction2DLattice(
        ba.HexagonalLattice2D(lattice_length, 0))
    pdf = ba.FTDecayFunction2DCauchy(10*nm, 10*nm, 0)
    interference.setDecayFunction(pdf)

    particle_layout.setInterferenceFunction(interference)

    vacuum_layer = ba.Layer(m_vacuum)
    vacuum_layer.addLayout(particle_layout)
    substrate_layer = ba.Layer(m_substrate, 0)
    multi_layer = ba.MultiLayer()
    multi_layer.addLayer(vacuum_layer)
    multi_layer.addLayer(substrate_layer)
    return multi_layer


def get_simulation(params):
    """
    Create and return GISAXS simulation with beam and detector defined
    """
    simulation = ba.GISASSimulation()
    simulation.setDetectorParameters(100, -1*deg, 1*deg, 100, 0, 2*deg)
    simulation.setBeamParameters(1*angstrom, 0.2*deg, 0)
    simulation.beam().setIntensity(1e+08)
    simulation.setSample(get_sample(params))
    return simulation


def create_real_data():
    """
    Generating "real" data by adding noise to the simulated data.
    """
    params = {'radius': 6*nm, 'length': 12*nm}
    simulation = get_simulation(params)
    simulation.runSimulation()

    # retrieving simulated data in the form of numpy array
    real_data = simulation.result().array()

    # spoiling simulated data with noise to produce "real" data
    np.random.seed(0)
    noise_factor = 0.1
    noisy = np.random.normal(real_data, noise_factor*np.sqrt(real_data))
    noisy[noisy < 0.1] = 0.1
    return noisy


class LMFITPlotter:
    """
    Adapts standard plotter for lmfit minimizer.
    """
    def __init__(self, fit_objective, every_nth=10):
        self.fit_objective = fit_objective
        self.plotter_gisas = ba_fitmonitor.PlotterGISAS()
        self.every_nth = every_nth

    def __call__(self, params, iter, resid):
        if iter % self.every_nth == 0:
            self.plotter_gisas.plot(self.fit_objective)


def run_fitting():
    """
    main function to run fitting
    """
    real_data = create_real_data()

    fit_objective = ba.FitObjective()
    fit_objective.addSimulationAndData(get_simulation, real_data, 1)
    fit_objective.initPrint(10)

    params = lmfit.Parameters()
    params.add('radius', value=7*nm, min=5*nm, max=8*nm)
    params.add('length', value=10*nm, min=8*nm, max=14*nm)

    plotter = LMFITPlotter(fit_objective)
    result = lmfit.minimize(fit_objective.evaluate_residuals,
                            params,
                            iter_cb=plotter)
    fit_objective.finalize(result)

    result.params.pretty_print()
    print(lmfit.fit_report(result))


if __name__ == '__main__':
    run_fitting()
    plt.show()
